import os

from PIL import Image, UnidentifiedImageError

from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.utils import get_mime_type
from deepsearchai.vector_databases.base import BaseVectorDatabase
from .base import BaseSource
from .data_source import DataSource


class LocalDataSource(BaseSource):
    def __init__(self):
        super().__init__()

    def add_data(
        self,
        source: str,
        embedding_models_config: EmbeddingModelsConfig,
        vector_database: BaseVectorDatabase,
    ) -> None:
        # Recursively iterate over all the files and subdirectories in the current directory
        existing_document_identifiers = {}
        file_paths = self._get_all_file_path(source)
        for file in file_paths:
            media_type = get_mime_type(file)
            embedding_models = embedding_models_config.get_embedding_model(media_type)
            for embedding_model in embedding_models:
                if media_type not in existing_document_identifiers:
                    existing_document_identifiers[
                        media_type
                    ] = vector_database.get_existing_document_ids(
                        {"document_id": file_paths},
                        embedding_model.get_collection_name(media_type),
                    )

                if file in existing_document_identifiers[media_type]:
                    "{} already exists, skipping...".format(file)
                    continue
                if media_type == MEDIA_TYPE.IMAGE:
                    try:
                        data = Image.open(file)
                    except FileNotFoundError:
                        print("The supplied file does not exist {}".format(file))
                        continue
                    except UnidentifiedImageError:
                        print("The supplied file is not an image {}".format(file))
                        continue
                    except Exception as e:
                        print("Error while reading file {}".format(file))
                        print(e)
                        continue

                elif media_type == MEDIA_TYPE.AUDIO:
                    data = file
                else:
                    print("Unsupported media type {}".format(file))
                    continue
                vector_database.add(
                    data, DataSource.LOCAL, file, source, media_type, embedding_model
                )

    def _get_all_file_path(self, directory):
        if os.path.isfile(directory):
            return [directory]

        file_paths = []
        for root, _, files in os.walk(directory):
            for file in files:
                file_path = os.path.join(root, file)
                file_paths.append(file_path)
        return file_paths



import copy
from typing import Any, Dict, List, Union

from deepsearchai.embedding_models.base import BaseEmbeddingModel
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from deepsearchai.types import MediaData
from .configs.base import BaseVectorDatabaseConfig


class BaseVectorDatabase:
    def __init__(self, config: BaseVectorDatabaseConfig):
        self.config = config
        pass

    def add(
        self,
        data: Any,
        datasource: DataSource,
        file: str,
        source: str,
        media_type: MEDIA_TYPE,
        embedding_model: BaseEmbeddingModel,
    ):
        raise NotImplementedError

    def query(
        self,
        query: str,
        n_results: int,
        media_type: MEDIA_TYPE,
        distance_threshold: float,
        embedding_model: BaseEmbeddingModel,
    ) -> List[MediaData]:
        raise NotImplementedError

    def get_existing_document_ids(
        self, document_ids: List[str], collection_name: str
    ) -> List[str]:
        raise NotImplementedError

    def _construct_metadata(
        self, metadata: List[Dict[str, Any]], source: str, document_id: str, len: int
    ):
        new_metadata = copy.deepcopy(metadata)
        is_metadata_empty = not metadata
        if is_metadata_empty:
            new_metadata = []

        for i in range(len):
            temp_metadata = {
                "source_type": DataSource.LOCAL.name,
                "source_id": source,
                "document_id": document_id,
            }
            if is_metadata_empty:
                new_metadata.append(temp_metadata)
            else:
                new_metadata[i].update(temp_metadata)

        return new_metadata



import io
import os
import uuid
from typing import Any

import openai
from pydub import AudioSegment

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from .base import BaseEmbeddingModel


# This is a model to transcribe audio to text using openai APIs, hence user needs to have the approrpiate env variables
# set to be able to use it
class WhisperOpenAi(BaseEmbeddingModel):
    MODEL_NAME = "whisper-1"
    SUPPORTED_MEDIA_TYPES = [MEDIA_TYPE.AUDIO, MEDIA_TYPE.VIDEO]

    def __init__(self):
        pass

    def get_media_encoding(
        self, file: str, data_type: MEDIA_TYPE, datasource: DataSource
    ):
        """Get the media encoding using OpenAI's Whisper model.

        Args:
            data: The media data.
            dataType: The media type, either "audio" or "video".

        Returns:
            The media encoding, or None if the encoding could not be determined.
            :param data:
            :param data_type:
        """
        if data_type not in self.SUPPORTED_MEDIA_TYPES:
            raise ValueError(
                "Unsupported dataType. Whisper model supports only {}".format(
                    self.SUPPORTED_MEDIA_TYPES
                )
            )
        audio = AudioSegment.from_file(file)
        chunk_size = self._get_chunk_size(audio.duration_seconds)
        # Split the AudioSegment file into chunks
        chunks = audio[::chunk_size]
        documents = []
        metadata = []
        ids = []
        for i, chunk in enumerate(chunks):
            # Transcribe the audio chunk
            buffer = io.BytesIO()
            buffer.name = ".tmp/deepsearch/youtube/tmp.wav"
            chunk.export(buffer, format="wav")

            transcript = openai.Audio.transcribe(self.MODEL_NAME, buffer)
            start = i * chunk_size
            end = start + chunk.duration_seconds
            documents.append(transcript.get("text"))
            metadata.append(
                {
                    "start": start,
                    "end": end,
                }
            )
            ids.append(str(uuid.uuid4()))

        if datasource != DataSource.LOCAL:
            self._delete_file(file)
        result = {"documents": documents, "metadata": metadata, "ids": ids}
        return result

    def get_text_encoding(self, query: str):
        return {"text": query}

    def get_collection_name(self, media_type: MEDIA_TYPE):
        return "deepsearch-{}".format(media_type.name.lower())

    def _get_chunk_size(self, total_duration: float):
        # Hardcoded chunk size of 5 minutes, but can have a smarter approach based on the total length of the audio
        return 300000

    def _delete_file(self, filename):
        os.remove(filename)



import logging
from typing import Optional

from .base import BaseVectorDatabaseConfig, EmbeddingFunction

try:
    import chromadb
    from chromadb.config import Settings
    from chromadb.errors import InvalidDimensionException
    from chromadb.utils.embedding_functions import \
        SentenceTransformerEmbeddingFunction

except RuntimeError:
    pass


class ChromaDbConfig(BaseVectorDatabaseConfig):
    def __init__(
        self,
        dir: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[str] = None,
        allow_reset=False,
        chroma_settings: Optional[dict] = None,
        embedding_function: Optional[EmbeddingFunction] = None,
    ):
        """
        Initializes a configuration class instance for ChromaDB.

        :param dir: Path to the database directory, where the database is stored, defaults to None
        :type dir: Optional[str], optional
        :param host: Database connection remote host. Use this if you run Embedchain as a client, defaults to None
        :type host: Optional[str], optional
        :param port: Database connection remote port. Use this if you run Embedchain as a client, defaults to None
        :type port: Optional[str], optional
        :param allow_reset: Resets the database. defaults to False
        :type allow_reset: bool
        :param chroma_settings: Chroma settings dict, defaults to None
        :type chroma_settings: Optional[dict], optional
        """

        self.settings = Settings()
        self.settings.allow_reset = allow_reset
        self.embedding_function = (
            SentenceTransformerEmbeddingFunction()
            if not embedding_function
            else embedding_function
        )
        if chroma_settings:
            for key, value in chroma_settings.items():
                if hasattr(self.settings, key):
                    setattr(self.settings, key, value)

        if host and port:
            logging.info(f"Connecting to ChromaDB server: {host}:{port}")
            self.settings.chroma_server_host = host
            self.settings.chroma_server_http_port = port
            self.settings.chroma_api_impl = "chromadb.api.fastapi.FastAPI"
        else:
            if dir is None:
                dir = "db"

            self.settings.persist_directory = dir
            self.settings.is_persistent = True

        super().__init__()



import uuid
from typing import Any

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from .base import BaseEmbeddingModel


class BlipImageCaptioning(BaseEmbeddingModel):
    MODEL_NAME = "Salesforce/blip-image-captioning-base"

    def __init__(self):
        self.processor = None
        self.model = None

    def get_media_encoding(
        self, data: Any, data_type: MEDIA_TYPE, datasource: DataSource
    ):
        self._load_model()
        inputs = self.processor(data, return_tensors="pt")
        out = self.model.generate(**inputs)
        caption = self.processor.decode(out[0], skip_special_tokens=True)
        id = str(uuid.uuid4())
        return {"documents": [caption], "ids": [id], "metadata": [{"type": "caption"}]}

    def get_text_encoding(self, query: str):
        return {"text": query}

    def get_collection_name(self, media_type: MEDIA_TYPE):
        return "deepsearch-{}-captioning".format(media_type.name.lower())

    def _load_model(self):
        if not self.processor or not self.model:
            try:
                from transformers import (BlipForConditionalGeneration,
                                          BlipProcessor)

                self.processor = BlipProcessor.from_pretrained(self.MODEL_NAME)
                self.model = BlipForConditionalGeneration.from_pretrained(
                    self.MODEL_NAME
                )
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "The required dependencies for audio/video are not installed."
                    ' Please install with `pip install --upgrade "deepsearchai[image]"`'
                )



from typing import Any, Dict, List, Optional

from deepsearchai.embedding_models.base import BaseEmbeddingModel
from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from deepsearchai.types import MediaData
from .base import BaseVectorDatabase
from .configs.chromadb import ChromaDbConfig

try:
    import chromadb
    from chromadb import Collection, QueryResult
    from chromadb.config import Settings
    from chromadb.errors import InvalidDimensionException

except ImportError:
    raise ImportError("Chromadb requires extra dependencies") from None


class ChromaDB(BaseVectorDatabase):
    """Vector database using ChromaDB."""

    BATCH_SIZE = 100

    def __init__(
        self,
        embedding_models_config: EmbeddingModelsConfig = EmbeddingModelsConfig(),
        config: Optional[ChromaDbConfig] = None,
    ):
        """Initialize a new ChromaDB instance

        :param config: Configuration options for Chroma, defaults to None
        :type config: Optional[ChromaDbConfig], optional
        """
        if config:
            self.config = config
        else:
            self.config = ChromaDbConfig()

        self.client = chromadb.Client(self.config.settings)
        self.embedding_models_config = embedding_models_config
        self._set_all_collections()
        super().__init__(config=self.config)

    def add(
        self,
        data: Any,
        datasource: DataSource,
        file: str,
        source: str,
        media_type: MEDIA_TYPE,
        embedding_model: BaseEmbeddingModel,
    ):
        encodings_json = embedding_model.get_media_encoding(
            data, media_type, datasource
        )
        embeddings = encodings_json.get("embedding", None)
        documents = (
            [file]
            if not encodings_json.get("documents")
            else encodings_json.get("documents")
        )
        metadata = self._construct_metadata(
            encodings_json.get("metadata", None), source, file, len(documents)
        )
        ids = encodings_json.get("ids", [])
        size = len(documents)
        if embeddings is not None and len(embeddings) != size:
            raise ValueError(
                "Cannot add documents to chromadb with inconsistent embeddings"
            )
        collection = self._get_or_create_collection(
            embedding_model.get_collection_name(media_type)
        )
        # embedding would be created by the llm model used
        for i in range(0, len(documents), self.BATCH_SIZE):
            print(
                "Inserting batches from {} to {} in chromadb".format(
                    i, min(len(documents), i + self.BATCH_SIZE)
                )
            )
            if embeddings is not None:
                collection.add(
                    embeddings=embeddings[i : i + self.BATCH_SIZE],
                    documents=documents[i : i + self.BATCH_SIZE],
                    ids=ids[i : i + self.BATCH_SIZE],
                    metadatas=metadata[i : i + self.BATCH_SIZE],
                )

            else:
                collection.add(
                    documents=documents[i : i + self.BATCH_SIZE],
                    ids=ids[i : i + self.BATCH_SIZE],
                    metadatas=metadata[i : i + self.BATCH_SIZE],
                )

    def query(
        self,
        query: str,
        n_results: int,
        media_type: MEDIA_TYPE,
        distance_threshold: float,
        embedding_model: BaseEmbeddingModel,
    ) -> List[MediaData]:
        response = embedding_model.get_text_encoding(query)
        input_embeddings = response.get("embedding", None)
        input_query = response.get("text", None)
        if input_embeddings:
            query_params = {
                "query_embeddings": [input_embeddings],
                "n_results": n_results,
            }
        else:
            query_params = {"query_texts": [input_query], "n_results": n_results}

        media_data = []

        collection = self._get_or_create_collection(
            embedding_model.get_collection_name(media_type)
        )
        try:
            results = collection.query(**query_params)
        except InvalidDimensionException as e:
            raise InvalidDimensionException(
                e.message()
                + ". This is commonly a side-effect when an embedding function, different from the one used to"
                " add the embeddings, is used to retrieve an embedding from the database."
            ) from None
        filtered_results = self.filter_query_result_by_distance(
            results, distance_threshold
        )
        if len(filtered_results.get("documents", [])) == 0:
            return media_data

        documents = filtered_results.get("documents")[0]
        metadatas = filtered_results.get("metadatas")[0]
        for document, metadata in zip(documents, metadatas):
            media_data.append({"document": document, "metadata": metadata})
        return media_data

    def filter_query_result_by_distance(
        self, query_result: QueryResult, distance_threshold: float
    ) -> QueryResult:
        filtered_result: QueryResult = {
            "ids": [],
            "embeddings": [],
            "documents": [],
            "metadatas": [],
            "distances": [],
        }

        for i, ids in enumerate(query_result["ids"]):
            filtered_subresult = {
                "ids": [],
                "embeddings": [],
                "documents": [],
                "metadatas": [],
                "distances": [],
            }
            if query_result["distances"][i] is None:
                continue

            for j, distance in enumerate(query_result["distances"][i]):
                if distance >= distance_threshold:
                    filtered_subresult["ids"].append(query_result["ids"][i][j])

                    if "embeddings" in query_result and query_result["embeddings"]:
                        embeddings = query_result["embeddings"][i]
                        filtered_subresult["embeddings"].append(embeddings[j])

                    if "documents" in query_result and query_result["documents"]:
                        documents = query_result["documents"][i]
                        filtered_subresult["documents"].append(documents[j])

                    if "metadatas" in query_result and query_result["metadatas"]:
                        metadatas = query_result["metadatas"][i]
                        filtered_subresult["metadatas"].append(metadatas[j])

                    filtered_subresult["distances"].append(distance)

            if filtered_subresult["ids"]:
                filtered_result["ids"].append(filtered_subresult["ids"])
                filtered_result["distances"].append(filtered_subresult["distances"])

                if filtered_subresult["embeddings"]:
                    filtered_result["embeddings"].append(
                        filtered_subresult["embeddings"]
                    )

                if filtered_subresult["documents"]:
                    filtered_result["documents"].append(filtered_subresult["documents"])

                if filtered_subresult["metadatas"]:
                    filtered_result["metadatas"].append(filtered_subresult["metadatas"])

        return filtered_result

    def get_existing_document_ids(
        self, metadata_filters, collection_name: str
    ) -> List[str]:
        query_args = {"where": self._generate_where_clause(metadata_filters)}
        collection = self._get_or_create_collection(collection_name)

        results = []
        offset = 0
        first_iteration = True
        while offset != -1 or first_iteration:
            first_iteration = False
            query_result = collection.get(
                **query_args, offset=offset, limit=self.BATCH_SIZE
            )
            metadatas = query_result.get("metadatas", [])
            document_ids = list(
                map(lambda metadata: metadata.get("document_id", []), metadatas)
            )
            results.extend(document_ids)
            offset = offset + min(self.BATCH_SIZE, len(query_result.get("ids")))
            if len(query_result.get("ids")) == 0:
                break
        return results

    def count(self) -> Dict[str, int]:
        """
        Count number of documents/chunks embedded in the database.

        :return: number of documents
        """
        return self._get_collection_count()

    def delete(self, where, media_type: Optional[MEDIA_TYPE] = None):
        if not media_type or media_type == MEDIA_TYPE.AUDIO:
            media_collections = self.collections.get(MEDIA_TYPE.AUDIO)
            for collection in media_collections:
                collection.delete(where=where)
        if not media_type or media_type == MEDIA_TYPE.IMAGE:
            media_collections = self.collections.get(MEDIA_TYPE.IMAGE)
            for collection in media_collections:
                collection.delete(where=where)
        if not media_type or media_type == MEDIA_TYPE.VIDEO:
            media_collections = self.collections.get(MEDIA_TYPE.VIDEO)
            for collection in media_collections:
                collection.delete(where=where)

    def reset(self):
        """
        Resets the database. Deletes all embeddings irreversibly.
        """
        # Delete all data from the collection
        for media_collections in self.collections.values():
            for collection in media_collections:
                try:
                    self.client.delete_collection(collection.name)
                except ValueError:
                    raise ValueError(
                        "For safety reasons, resetting is disabled. "
                        "Please enable it by setting `allow_reset=True` in your ChromaDbConfig"
                    ) from None
        self._set_all_collections()

    def _get_or_create_collection(
        self,
        collection_name: str,
    ):
        return self.client.get_or_create_collection(
            name=collection_name,
            embedding_function=self.config.embedding_function,
            metadata={"hnsw:space": "cosine"},
        )

    def _generate_where_clause(self, where_clause: Dict[str, any]):
        # If only one filter is supplied, return it as is
        # (no need to wrap in $and based on chroma docs)
        if not where_clause:
            return {}
        if len(where_clause.keys()) == 1:
            value = list(where_clause.values())[0]
            key = list(where_clause.keys())[0]
            if isinstance(value, list):
                where_filter = {key: {"$in": value}}
            else:
                where_filter = {key: value}
            return where_filter
        where_filters = []
        for k, v in where_clause.items():
            if isinstance(v, list):
                where_filters.append({k: {"$in": v}})
            if isinstance(v, str):
                where_filters.append({k: v})
        return {"$and": where_filters}

    def _set_all_collections(self):
        collections = {}
        for item in self.embedding_models_config.llm_models.items():
            collections[item[0]] = []
            for embedding_model in item[1]:
                media_type = item[0]
                collection_name = embedding_model.get_collection_name(media_type)
                collections[media_type].append(
                    self._get_or_create_collection(collection_name)
                )
        self.collections = collections

    def _get_collection_count(self):
        collections_count = {}
        for media_collections in self.collections.values():
            for collection in media_collections:
                collections_count[collection.name] = collection.count()
        return collections_count



from typing import Dict, List

from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.llms.base import BaseLLM
from deepsearchai.llms.openai import OpenAi
from deepsearchai.sources.utils import SourceUtils
from deepsearchai.types import MediaData, QueryResult
from deepsearchai.vector_databases.base import BaseVectorDatabase
from deepsearchai.vector_databases.chromadb import ChromaDB


class App:
    def __init__(
        self,
        embedding_models_config: EmbeddingModelsConfig,
        vector_database: BaseVectorDatabase,
        llm: BaseLLM,
    ):
        self.embedding_models_config = (
            embedding_models_config
            if embedding_models_config
            else EmbeddingModelsConfig()
        )
        self.vector_database = (
            vector_database
            if vector_database
            else ChromaDB(embedding_models_config=self.embedding_models_config)
        )

        self.llm = llm if llm else OpenAi(self.vector_database)
        self.source_utils = SourceUtils()

    def add_data(self, source: str):
        self.source_utils.add_data(
            source, self.embedding_models_config, self.vector_database
        )

    def query(
        self, query: str, media_types: List[MEDIA_TYPE] = [MEDIA_TYPE.IMAGE]
    ) -> QueryResult:
        data = self.get_data(query, media_types)
        response = self.llm.query(query, data)
        return response

    def get_data(
        self, query: str, media_types: List[MEDIA_TYPE] = [MEDIA_TYPE.IMAGE]
    ) -> Dict[MEDIA_TYPE, List[MediaData]]:
        return self.source_utils.get_data(
            query, media_types, self.embedding_models_config, self.vector_database
        )



from typing import Any

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource


class BaseEmbeddingModel:
    def __init__(self):
        pass

    def get_media_encoding(
        self, data: Any, dataType: MEDIA_TYPE, datasource: DataSource
    ):
        raise NotImplementedError

    def get_text_encoding(self, query: str):
        raise NotImplementedError

    def get_collection_name(self, media_type: MEDIA_TYPE):
        raise NotImplementedError



import unittest
from unittest import mock
from unittest.mock import patch

from deepsearchai.embedding_models.blip_image_captioning import \
    BlipImageCaptioning
from deepsearchai.embedding_models.clip import Clip
from deepsearchai.embedding_models.whisper_openai import WhisperOpenAi
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from deepsearchai.vector_databases.chromadb import ChromaDB
from deepsearchai.vector_databases.configs.chromadb import ChromaDbConfig


class ChromaDBTest(unittest.TestCase):
    blip_image_captioning = BlipImageCaptioning()
    whisper_openai = WhisperOpenAi()
    clip = Clip()

    @patch("chromadb.Client")
    def test_init(self, chromadb_client):
        config = ChromaDbConfig()
        chromadb = ChromaDB(config=config)

        self.assertEqual(chromadb.config, config)
        self.assertEqual(chromadb.client, chromadb_client.return_value)
        self.assertEqual(
            chromadb_client.return_value.get_or_create_collection.mock_calls,
            [
                mock.call(
                    name=self.whisper_openai.get_collection_name(MEDIA_TYPE.AUDIO),
                    embedding_function=config.embedding_function,
                    metadata={"hnsw:space": "cosine"},
                ),
                mock.call(
                    name=self.clip.get_collection_name(MEDIA_TYPE.IMAGE),
                    embedding_function=config.embedding_function,
                    metadata={"hnsw:space": "cosine"},
                ),
                mock.call(
                    name=self.blip_image_captioning.get_collection_name(
                        MEDIA_TYPE.IMAGE
                    ),
                    embedding_function=config.embedding_function,
                    metadata={"hnsw:space": "cosine"},
                ),
                mock.call(
                    name=self.whisper_openai.get_collection_name(MEDIA_TYPE.VIDEO),
                    embedding_function=config.embedding_function,
                    metadata={"hnsw:space": "cosine"},
                ),
            ],
        )

    @patch("chromadb.Client")
    def test_add(self, chromadb_client):
        # Mock clip model to be able to mock the corresponding generated embeddings
        clip_model_mock = mock.Mock()
        clip_model_mock.get_collection_name.return_value = (
            self.clip.get_collection_name(MEDIA_TYPE.IMAGE)
        )

        # Mock embedding models config, and make it return the mocked clip model
        embedding_models_config = mock.Mock()
        embedding_models_config.llm_models.items.return_value = [
            (MEDIA_TYPE.IMAGE, [clip_model_mock])
        ]

        # Initialize chromadb with the mocks
        config = ChromaDbConfig()
        chromadb = ChromaDB(
            embedding_models_config=embedding_models_config, config=config
        )

        mock_image_collection = chromadb_client.return_value.get_or_create_collection(
            name=self.clip.get_collection_name(MEDIA_TYPE.IMAGE),
            embedding_function=config.embedding_function,
            metadata={"hnsw:space": "cosine"},
        )

        embeddings = [[1.0, 2.0]]
        documents = ["file"]
        ids = ["id1"]
        metadatas = [
            {"source_type": "LOCAL", "source_id": "source", "document_id": "file"}
        ]

        clip_model_mock.get_media_encoding.return_value = {
            "embedding": embeddings,
            "ids": ids,
        }

        chromadb.add(
            "local_file_path",
            DataSource.LOCAL,
            "file",
            "source",
            MEDIA_TYPE.IMAGE,
            clip_model_mock,
        )

        mock_image_collection.add.assert_called_once()
        self.assertEqual(
            mock_image_collection.add.mock_calls,
            [
                mock.call(
                    embeddings=embeddings,
                    documents=documents,
                    ids=ids,
                    metadatas=metadatas,
                )
            ],
        )

    @patch("chromadb.Client")
    def test_query(self, chromadb_client):
        # Mock clip model to be able to mock the corresponding generated embeddings
        clip_model_mock = mock.Mock()
        clip_model_mock.get_collection_name.return_value = (
            self.clip.get_collection_name(MEDIA_TYPE.IMAGE)
        )

        # Mock embedding models config, and make it return the mocked clip model
        embedding_models_config = mock.Mock()
        embedding_models_config.llm_models.items.return_value = [
            (MEDIA_TYPE.IMAGE, [clip_model_mock])
        ]

        # Initialize chromadb with the mocks
        config = ChromaDbConfig()
        chromadb = ChromaDB(
            embedding_models_config=embedding_models_config, config=config
        )

        mock_image_collection = chromadb_client.return_value.get_or_create_collection(
            name=self.clip.get_collection_name(MEDIA_TYPE.IMAGE),
            embedding_function=config.embedding_function,
            metadata={"hnsw:space": "cosine"},
        )

        mock_image_collection.query.return_value = {
            "ids": [["document_image_id1", "document_image_id2"]],
            "documents": [
                [
                    "This is image document 1",
                    "This is image document 2",
                ],
            ],
            "metadatas": [[{"source": "imagesource1"}, {"source": "imagesource2"}]],
            "distances": [[0.721456789, 0.456789012]],
        }

        input_query = "This is a query"
        input_embeddings = [1.0, 2.0, 3.0]
        n_results = 10

        clip_model_mock.get_text_encodding.return_value = {
            "embedding": input_embeddings,
        }

        results = chromadb.query(
            input_query, n_results, MEDIA_TYPE.IMAGE, 0.5, clip_model_mock
        )
        self.assertEqual(
            results,
            [
                {
                    "document": "This is image document 1",
                    "metadata": {"source": "imagesource1"},
                }
            ],
        )

    @patch("chromadb.Client")
    def test_get_existing_document_ids(self, chromadb_client):
        # Mock clip model to be able to mock the corresponding generated embeddings
        collection_name = self.clip.get_collection_name(MEDIA_TYPE.IMAGE)
        clip_model_mock = mock.Mock()
        clip_model_mock.get_collection_name.return_value = collection_name

        # Mock embedding models config, and make it return the mocked clip model
        embedding_models_config = mock.Mock()
        embedding_models_config.llm_models.items.return_value = [
            (MEDIA_TYPE.IMAGE, [clip_model_mock])
        ]

        # Initialize chromadb with the mocks
        config = ChromaDbConfig()
        chromadb = ChromaDB(
            embedding_models_config=embedding_models_config, config=config
        )

        mock_image_collection = chromadb_client.return_value.get_or_create_collection(
            name=self.clip.get_collection_name(MEDIA_TYPE.IMAGE),
            embedding_function=config.embedding_function,
            metadata={"hnsw:space": "cosine"},
        )

        mock_image_collection.get.side_effect = [
            {
                "ids": ["id1"],
                "metadatas": [{"document_id": "document1"}],
            },
            {
                "ids": [],
                "metadatas": [],
            },
        ]

        metadata_filters = {"key1": "value1"}
        self.assertEqual(
            chromadb.get_existing_document_ids(metadata_filters, collection_name),
            ["document1"],
        )

    @patch("chromadb.Client")
    def test_count(self, chromadb_client):
        mock_image_collection = mock.Mock()
        mock_image_caption_collection = mock.Mock()
        mock_audio_collection = mock.Mock()
        mock_video_collection = mock.Mock()
        chromadb_client.return_value.get_or_create_collection.side_effect = [
            mock_audio_collection,
            mock_image_collection,
            mock_image_caption_collection,
            mock_video_collection,
        ]

        config = ChromaDbConfig()
        chromadb = ChromaDB(config=config)

        mock_image_collection.count.return_value = 100
        mock_image_caption_collection.count.return_value = 150
        mock_audio_collection.count.return_value = 200
        mock_video_collection.count.return_value = 300

        count = chromadb.count()

        assert count == {
            mock_image_collection.name: 100,
            mock_image_caption_collection.name: 150,
            mock_audio_collection.name: 200,
            mock_video_collection.name: 300,
        }

    @patch("chromadb.Client")
    def test_delete(self, chromadb_client):
        mock_image_collection = mock.Mock()
        mock_image_caption_collection = mock.Mock()
        mock_audio_collection = mock.Mock()
        mock_video_collection = mock.Mock()
        chromadb_client.return_value.get_or_create_collection.side_effect = [
            mock_audio_collection,
            mock_image_collection,
            mock_image_caption_collection,
            mock_video_collection,
        ]

        config = ChromaDbConfig()
        chromadb = ChromaDB(config=config)

        chromadb.delete({})

        mock_image_collection.delete.assert_called_once()
        mock_image_caption_collection.delete.assert_called_once()
        mock_audio_collection.delete.assert_called_once()
        mock_video_collection.delete.assert_called_once()

        mock_image_collection.reset_mock()
        mock_image_caption_collection.reset_mock()
        mock_audio_collection.reset_mock()
        mock_video_collection.reset_mock()

        chromadb.delete({}, media_type=MEDIA_TYPE.IMAGE)

        mock_image_collection.delete.assert_called_once()
        mock_image_caption_collection.delete.assert_called_once()
        mock_audio_collection.delete.assert_not_called()
        mock_video_collection.delete.assert_not_called()

        mock_image_collection.reset_mock()
        mock_image_caption_collection.reset_mock()
        mock_audio_collection.reset_mock()
        mock_video_collection.reset_mock()

        chromadb.delete({}, media_type=MEDIA_TYPE.AUDIO)

        mock_audio_collection.delete.assert_called_once()
        mock_image_collection.delete.assert_not_called()
        mock_image_caption_collection.delete.assert_not_called()
        mock_video_collection.delete.assert_not_called()

        mock_image_collection.reset_mock()
        mock_image_caption_collection.reset_mock()
        mock_audio_collection.reset_mock()
        mock_video_collection.reset_mock()

        chromadb.delete({}, media_type=MEDIA_TYPE.VIDEO)

        mock_video_collection.delete.assert_called_once()
        mock_image_collection.delete.assert_not_called()
        mock_image_caption_collection.delete.assert_not_called()
        mock_audio_collection.delete.assert_not_called()



import uuid
from typing import Any

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from .base import BaseEmbeddingModel


class Whisper(BaseEmbeddingModel):
    MODEL_NAME = "base"
    SUPPORTED_MEDIA_TYPES = [MEDIA_TYPE.AUDIO, MEDIA_TYPE.VIDEO]

    def __init__(self):
        self.model = None

    def get_media_encoding(
        self, data: Any, data_type: MEDIA_TYPE, datasource: DataSource
    ):
        """Get the media encoding using OpenAI's Whisper model.

        Args:
            data: The media data.
            dataType: The media type, either "audio" or "video".

        Returns:
            The media encoding, or None if the encoding could not be determined.
            :param data:
            :param data_type:
        """

        if data_type not in self.SUPPORTED_MEDIA_TYPES:
            raise ValueError(
                "Unsupported dataType. Whisper model supports only {}".format(
                    self.SUPPORTED_MEDIA_TYPES
                )
            )
        self._load_model()
        transcription = self.model.transcribe(data)
        documents = []
        metadata = []
        ids = []
        for segment in transcription.get("segments"):
            documents.append(segment.get("text"))
            metadata.append(
                {
                    "start": segment.get("start"),
                    "end": segment.get("end"),
                }
            )
            ids.append(str(uuid.uuid4()))

        result = {"documents": documents, "metadata": metadata, "ids": ids}
        return result

    def get_text_encoding(self, query: str):
        return {"text": query}

    def get_collection_name(self, media_type: MEDIA_TYPE):
        return "deepsearch-{}".format(media_type.name.lower())

    def _load_model(self):
        if not self.model:
            try:
                import whisper

                # Load the Whisper Model
                self.model = whisper.load_model(self.MODEL_NAME)
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "The required dependencies for audio/video are not installed."
                    ' Please install with `pip install --upgrade "deepsearchai[audio]"` '
                    'or `pip install --upgrade "deepsearchai[video]"`'
                )



from typing import Dict, List, Optional

from typing_extensions import TypedDict

from deepsearchai.enums import MEDIA_TYPE


class MediaData(TypedDict):
    document: str
    metadata: Optional[Dict[str, str]]


class QueryResult(TypedDict):
    llm_response: str
    documents: Optional[Dict[MEDIA_TYPE, List[MediaData]]]



from deepsearchai.embedding_models.base import BaseEmbeddingModel
from deepsearchai.vector_databases.base import BaseVectorDatabase


class BaseSource:
    def __init__(self):
        pass

    def add_data(
        self,
        source: str,
        llm_model: BaseEmbeddingModel,
        vector_database: BaseVectorDatabase,
    ) -> None:
        raise NotImplementedError



import io
import unittest
from unittest.mock import patch

import boto3
import mock
from PIL import Image

from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from deepsearchai.sources.s3 import S3DataSource
from deepsearchai.vector_databases.base import BaseVectorDatabase


class S3DataSourceTests(unittest.TestCase):
    @patch.object(boto3, "client")
    def setUp(self, mock_boto3_client):
        self.s3_data_source = S3DataSource()
        self.llms_config = EmbeddingModelsConfig()
        self.mock_s3_client = mock_boto3_client.return_value

    def create_fake_image_data(self):
        image = Image.new("RGB", (100, 100), (255, 0, 0))
        image_bytes = io.BytesIO()
        image.save(image_bytes, format="JPEG")
        image_bytes.seek(0)
        return image_bytes

    def test_load_image_from_s3(self):
        bucket_name = "my-bucket"
        object_key = "my-image-file.jpg"
        self.mock_s3_client.get_object.return_value = {
            "Body": self.create_fake_image_data()
        }
        image = self.s3_data_source._load_image_from_s3(bucket_name, object_key)
        self.assertIsInstance(image, Image.Image)

    def test_get_all_objects_inside_an_object(self):
        bucket_name = "my-bucket"
        object_key = "my-folder"
        self.mock_s3_client.list_objects_v2.return_value = {
            "Contents": [
                {"Key": "my-folder/my-file1.jpg"},
                {"Key": "my-folder/my-file2.jpg"},
            ]
        }
        files, s3_paths = self.s3_data_source._get_all_objects_inside_an_object(
            bucket_name, object_key
        )
        self.assertEqual(files, ["my-folder/my-file1.jpg", "my-folder/my-file2.jpg"])
        self.assertEqual(
            s3_paths,
            [
                "s3://my-bucket/my-folder/my-file1.jpg",
                "s3://my-bucket/my-folder/my-file2.jpg",
            ],
        )

    def test_get_s3_bucket_name(self):
        url = "s3://my-bucket/my-folder/my-file.jpg"
        bucket_name = self.s3_data_source._get_s3_bucket_name(url)
        self.assertEqual(bucket_name, "my-bucket")

    def test_get_s3_object_key_name(self):
        url = "s3://my-bucket/my-folder/my-file.jpg"
        object_key = self.s3_data_source._get_s3_object_key_name(url)
        self.assertEqual(object_key, "my-folder/my-file.jpg")

    def test_add_data_for_image(self):
        source = "s3://my-bucket/my-folder/my-image.jpg"
        mock_vector_database = mock.Mock(BaseVectorDatabase)
        mock_vector_database.get_existing_document_ids.return_value = {}

        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]

        # Mock the get_all_objects_inside_an_object method to return a single object
        with patch.object(
            S3DataSource, "_get_all_objects_inside_an_object"
        ) as mock_get_all_objects_inside_an_object:
            mock_get_all_objects_inside_an_object.return_value = (
                ["my-image.jpg"],
                ["s3://my-bucket/my-folder/my-image.jpg"],
            )

            # Mock the load_image_from_s3 method to return a valid image
            with patch.object(
                S3DataSource, "_load_image_from_s3"
            ) as mock_load_image_from_s3:
                image_data = Image.new("RGB", (100, 100), (255, 0, 0))
                mock_load_image_from_s3.return_value = image_data

                directory = "s3://my-bucket/my-folder/my-image.jpg"

                self.s3_data_source.add_data(
                    source, embedding_models_config, mock_vector_database
                )
                mock_vector_database.add.assert_called_once()
                args, kwargs = mock_vector_database.add.call_args
                self.assertEqual(args[0], image_data)
                self.assertEqual(args[1], DataSource.LOCAL)
                self.assertEqual(args[2], source)
                self.assertEqual(args[3], source)
                self.assertEqual(args[4], MEDIA_TYPE.IMAGE)
                self.assertEqual(args[5], embedding_model)

    def test_add_data_for_audio(self):
        source = "s3://my-bucket/my-folder/my-audio.mp3"
        bucket_name = "my-bucket"
        object_key = "my-folder/my-audio.mp3"
        mock_vector_database = mock.Mock(BaseVectorDatabase)
        mock_vector_database.get_existing_document_ids.return_value = {}

        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]

        # Mock the get_all_objects_inside_an_object method to return a single object
        with patch.object(
            S3DataSource, "_get_all_objects_inside_an_object"
        ) as mock_get_all_objects_inside_an_object:
            mock_get_all_objects_inside_an_object.return_value = (
                ["my-audio.mp3"],
                ["s3://my-bucket/my-folder/my-audio.mp3"],
            )

            # Mock the load_audio_from_s3 method to return a valid audio data
            with patch.object(
                S3DataSource, "_load_audio_from_s3"
            ) as mock_load_audio_from_s3:
                mock_load_audio_from_s3.return_value = "/tmp/deepsearch/my-audio.mp3"

                # Create a mock for the arguments to be received from the llm model
                embeddings = [[0, 0, 0]]
                metadata = [{"author": "author1"}]
                ids = ["id1"]
                encodings_json = {
                    "embedding": embeddings,
                    "metadata": metadata,
                    "ids": ids,
                }

                self.s3_data_source.add_data(
                    source, embedding_models_config, mock_vector_database
                )

        # Verify that the vector database add method was called with the correct arguments

        mock_vector_database.add.assert_called_once()
        args, kwargs = mock_vector_database.add.call_args

        self.assertEqual(args[0], "/tmp/deepsearch/my-audio.mp3")
        self.assertEqual(args[1], DataSource.LOCAL)
        self.assertEqual(args[2], source)
        self.assertEqual(args[3], source)
        self.assertEqual(args[4], MEDIA_TYPE.AUDIO)
        self.assertEqual(args[5], embedding_model)



from typing import List, Sequence, Union

Vector = Union[Sequence[float], Sequence[int]]
Embedding = Vector
Embeddings = List[Embedding]
Document = str
Documents = List[Document]


class EmbeddingFunction:
    def __call__(self, texts: Documents) -> Embeddings:
        ...


class BaseVectorDatabaseConfig:
    def __init__(self):
        pass



from typing import Optional

from deepsearchai.embedding_models.base import BaseEmbeddingModel
from deepsearchai.embedding_models.blip_image_captioning import BlipImageCaptioning
from deepsearchai.embedding_models.clip import Clip
from deepsearchai.embedding_models.whisper_model import Whisper
from deepsearchai.enums import MEDIA_TYPE


class EmbeddingModelsConfig:
    def __init__(
        self,
        image_embedding_model: Optional[BaseEmbeddingModel] = None,
        audio_embedding_model: Optional[BaseEmbeddingModel] = None,
        video_embedding_model: Optional[BaseEmbeddingModel] = None,
        image_captioning_model: Optional[BaseEmbeddingModel] = BlipImageCaptioning(),
    ):
        if not image_embedding_model:
            image_embedding_model = Clip()
        if not audio_embedding_model:
            audio_embedding_model = Whisper()
        if not video_embedding_model:
            video_embedding_model = Whisper()
        image_embedding_models = [image_embedding_model]
        audio_embedding_models = [audio_embedding_model]
        video_embedding_models = [video_embedding_model]
        if image_captioning_model:
            image_embedding_models.append(image_captioning_model)

        self.llm_models = {
            MEDIA_TYPE.AUDIO: audio_embedding_models,
            MEDIA_TYPE.IMAGE: image_embedding_models,
            MEDIA_TYPE.VIDEO: video_embedding_models,
        }

    def get_embedding_model(self, media_type: MEDIA_TYPE):
        return self.llm_models.get(media_type, [])



import mimetypes
import os
import re
from typing import Dict, List

from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.types import MediaData
from deepsearchai.vector_databases.base import BaseVectorDatabase
from .data_source import DataSource
from .local import LocalDataSource
from .s3 import S3DataSource
from .youtube import YoutubeDatasource


class SourceUtils:
    def __init__(self):
        self.local_data_source = LocalDataSource()
        self.s3_data_source = S3DataSource()
        self.youtube_data_source = YoutubeDatasource()

    def add_data(
        self,
        source: str,
        embedding_models_config: EmbeddingModelsConfig,
        vector_database: BaseVectorDatabase,
    ) -> None:
        datasource = self._infer_type(source)
        if datasource == DataSource.S3:
            self.s3_data_source.add_data(
                source, embedding_models_config, vector_database
            )
        elif datasource == DataSource.LOCAL:
            self.local_data_source.add_data(
                source, embedding_models_config, vector_database
            )
        elif datasource == DataSource.YOUTUBE:
            self.youtube_data_source.add_data(
                source, embedding_models_config, vector_database
            )
        else:
            raise ValueError("Invalid data source")

    def get_data(
        self,
        query: str,
        media_types: List[MEDIA_TYPE],
        embedding_models_config: EmbeddingModelsConfig,
        vector_database: BaseVectorDatabase,
    ) -> Dict[MEDIA_TYPE, List[MediaData]]:
        media_data = {}
        for media_type in media_types:
            if media_type == MEDIA_TYPE.UNKNOWN:
                continue
            media_data[media_type] = []
            for embedding_model in embedding_models_config.get_embedding_model(
                media_type
            ):
                media_data[media_type].extend(
                    vector_database.query(query, 1, media_type, 0.5, embedding_model)
                )
        return media_data

    def _infer_type(self, source: str) -> DataSource:
        if self._is_s3_path(source):
            return DataSource.S3
        elif self._is_local_datasource(source):
            return DataSource.LOCAL
        elif self._is_youtube_datasource(source):
            return DataSource.YOUTUBE
        else:
            raise ValueError("Invalid data source")

    def _is_s3_path(self, path: str):
        """Checks if a supplied string is an S3 path."""

        # Regex pattern for an S3 path
        # s3_path_regex = r'^s3://(?P<bucket>[A-Za-z0-9\-\.]+)/(?P<key>.*)$'
        s3_path_regex = r"^s3://(?P<bucket>[A-Za-z0-9\-\.\/]+)$"

        # Match the path against the regex pattern
        match = re.match(s3_path_regex, path)

        # If the path matches the regex pattern, then it is an S3 path
        return match is not None

    def _is_local_datasource(self, source: str) -> bool:
        """Checks if a supplied string is a local directory or a file.

        Args:
          source: The string to check.

        Returns:
          True if the string is a local directory or a file, False otherwise.
        """
        if os.path.isdir(source):
            return True
        elif os.path.isfile(source):
            return True
        else:
            return False

    def _is_youtube_datasource(self, source: str) -> bool:
        """Checks if a supplied string is a youtube channel id

        Args:
          source: The string to check.

        Returns:
          True if the string is a Youtube channel id
        """
        tokens = source.split(":")
        if len(tokens) == 2 and tokens[0] == "youtube":
            return True
        else:
            return False



import mimetypes

from deepsearchai.enums import MEDIA_TYPE


def get_mime_type(filename: str) -> MEDIA_TYPE:
    mime_type, encoding = mimetypes.guess_type(filename)
    if not mime_type or mime_type.split("/")[0].upper() not in MEDIA_TYPE.__members__:
        return MEDIA_TYPE.UNKNOWN
    return MEDIA_TYPE[mime_type.split("/")[0].upper()]



import io
import os
import urllib.parse

import boto3
from PIL import Image, UnidentifiedImageError

from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.base import BaseSource
from deepsearchai.sources.data_source import DataSource
from deepsearchai.utils import get_mime_type
from deepsearchai.vector_databases.base import BaseVectorDatabase


class S3DataSource(BaseSource):
    def __init__(self):
        self.access_key = os.environ.get("AWS_ACCESS_KEY")
        self.secret_key = os.environ.get("AWS_SECRET_KEY")
        self.client = boto3.client(
            "s3",
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            region_name="us-east-1",
        )
        super().__init__()

    def add_data(
        self,
        source: str,
        embedding_models_config: EmbeddingModelsConfig,
        vector_database: BaseVectorDatabase,
    ) -> None:
        bucket_name = self._get_s3_bucket_name(source)
        key = self._get_s3_object_key_name(source)
        objects, s3_paths = self._get_all_objects_inside_an_object(bucket_name, key)
        existing_document_identifiers = {}
        for s3_object, object_s3_path in zip(objects, s3_paths):
            media_type = get_mime_type(s3_object)
            embedding_models = embedding_models_config.get_embedding_model(media_type)
            for embedding_model in embedding_models:
                if media_type not in existing_document_identifiers:
                    existing_document_identifiers[
                        media_type
                    ] = vector_database.get_existing_document_ids(
                        {"document_id": s3_paths},
                        embedding_model.get_collection_name(media_type),
                    )

                if object_s3_path in existing_document_identifiers[media_type]:
                    "{} already exists, skipping...".format(object_s3_path)
                    continue
                if media_type == MEDIA_TYPE.IMAGE:
                    data = self._load_image_from_s3(bucket_name, s3_object)
                    if data is None:
                        continue
                elif media_type == MEDIA_TYPE.AUDIO:
                    data = self._load_audio_from_s3(bucket_name, s3_object)
                else:
                    print("Unsupported media type {}".format(s3_object))
                    continue
                vector_database.add(
                    data,
                    DataSource.LOCAL,
                    object_s3_path,
                    source,
                    media_type,
                    embedding_model,
                )

    def _load_audio_from_s3(self, bucket_name, object_key):
        """Loads an audio file from S3 and returns the audio data."""
        # Create the directory if it doesn't exist
        if not os.path.exists("/tmp/deepsearch"):
            os.makedirs("/tmp/deepsearch")
        local_file_path = "/tmp/deepsearch/{}".format(object_key)
        self.client.download_file(bucket_name, object_key, local_file_path)
        return local_file_path

    def _load_image_from_s3(self, bucket_name, object_key):
        """Loads an image from S3 and opens it using PIL.

        Args:
          bucket_name: The name of the S3 bucket.
          object_key: The key of the image object.

        Returns:
          A PIL Image object.
        """

        response = self.client.get_object(Bucket=bucket_name, Key=object_key)
        image_data = response["Body"].read()

        image_stream = io.BytesIO(image_data)
        try:
            return Image.open(image_stream)
        except UnidentifiedImageError:
            print(
                "The supplied file is not an image {}".format(
                    "{}/{}".format(bucket_name, object_key)
                )
            )
            return None
        except Exception as e:
            print(
                "Error while reading file {}".format(
                    "{}/{}".format(bucket_name, object_key)
                )
            )
            print(e)
            return None

    def _get_s3_bucket_name(self, url):
        """Extracts the S3 bucket name from its URL.

        Args:
          url: The S3 URL.

        Returns:
          The S3 bucket name.
        """

        parsed_url = urllib.parse.urlparse(url, allow_fragments=False)
        return parsed_url.netloc

    def _get_s3_object_key_name(self, url):
        """Gets the S3 object name from its URL.

        Args:
          url: The S3 URL.

        Returns:
          The S3 object name.
        """

        parsed_url = urllib.parse.urlparse(url, allow_fragments=False)
        return parsed_url.path.strip("/")

    def _get_all_objects_inside_an_object(self, bucket_name, object_key):
        """Lists all the files inside a folder in an S3 bucket, but does not add the sub-folders.

        Args:
          bucket_name: The name of the S3 bucket.
          object_key: The key of the folder to list the files inside.

        Returns:
          A list of the names of all the files inside the folder.
        """

        files = []
        s3_paths = []
        if not object_key:
            response = self.client.list_objects_v2(Bucket=bucket_name)
        else:
            response = self.client.list_objects_v2(
                Bucket=bucket_name, Prefix=object_key
            )

        while True:
            for object in response["Contents"]:
                if object["Key"].endswith("/"):
                    continue
                files.append(object["Key"])
                s3_paths.append("s3://{}/{}".format(bucket_name, object["Key"]))

            if "NextContinuationToken" in response:
                response = self.client.list_objects_v2(
                    Bucket=bucket_name,
                    Prefix=object_key + "/",
                    ContinuationToken=response["NextContinuationToken"],
                )
            else:
                break
        return files, s3_paths



import os

from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.vector_databases.base import BaseVectorDatabase
from .base import BaseSource
from .data_source import DataSource


class YoutubeDatasource(BaseSource):
    OUTPUT_PATH = "tmp/deepsearch/youtube/"

    def __init__(self):
        self.youtube_client = None
        super().__init__()

    def add_data(
        self,
        source: str,
        embedding_models_config: EmbeddingModelsConfig,
        vector_database: BaseVectorDatabase,
    ) -> None:
        self._set_youtube_client()
        channel_id = source.split(":")[1]
        video_ids = self._get_channel_video_ids(channel_id)
        for video_id in video_ids:
            data = self._chunk_and_load_video(video_id)
            embedding_models = embedding_models_config.get_embedding_model(
                MEDIA_TYPE.VIDEO
            )
            for embedding_model in embedding_models:
                vector_database.add(
                    data,
                    DataSource.LOCAL,
                    video_id,
                    source,
                    MEDIA_TYPE.VIDEO,
                    embedding_model,
                )

    def _chunk_and_load_video(self, video_id):
        try:
            # Download the audio of the video
            import pytube

            yt = pytube.YouTube(f"https://www.youtube.com/watch?v={video_id}")
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "The required dependencies for audio/video are not installed."
                ' Please install with `pip install --upgrade "deepsearchai[video]"`'
            )
        audio = yt.streams.filter(only_audio=True).first()
        filename = "{}/{}".format(self.OUTPUT_PATH, audio.default_filename)
        audio.download(output_path=self.OUTPUT_PATH)
        return filename

    def _get_channel_video_ids(self, channel_id):
        """Gets the video IDs for a YouTube channel.

        Args:
            channel_id: The ID of the YouTube channel.

        Returns:
            A list of video IDs for the YouTube channel.
        """
        channel_resource = (
            self.youtube_client.channels()
            .list(id=channel_id, part="contentDetails")
            .execute()
        )

        # Get the channel's upload playlist ID
        upload_playlist_id = channel_resource["items"][0]["contentDetails"][
            "relatedPlaylists"
        ]["uploads"]

        # Retrieve the playlist's video list
        playlist_items = (
            self.youtube_client.playlistItems()
            .list(part="snippet", playlistId=upload_playlist_id)
            .execute()
        )

        # Get the video IDs from the playlist's video items
        video_ids = []
        for item in playlist_items["items"]:
            video_ids.append(item["snippet"]["resourceId"]["videoId"])

        return video_ids

    def _set_youtube_client(self):
        # Create a YouTube API service object
        try:
            import googleapiclient.discovery

            if not self.youtube_client:
                self.youtube_client = googleapiclient.discovery.build(
                    "youtube",
                    "v3",
                    developerKey=os.environ.get("GOOGLE_CLIENT_API_KEY"),
                )
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "The required dependencies for audio/video are not installed."
                ' Please install with `pip install --upgrade "deepsearchai[video]"`'
            )



import unittest

from deepsearchai.app import App
from deepsearchai.embedding_models.clip import Clip
from deepsearchai.embedding_models_config import EmbeddingModelsConfig
from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.llms.openai import OpenAi
from deepsearchai.sources.utils import SourceUtils
from deepsearchai.vector_databases.chromadb import ChromaDB


class TestS3(unittest.TestCase):
    # TODO: Handle full file path as input
    def test_add_s3_and_get_data_image(self):
        ChromaDB().reset()
        app = App(None, None, None)
        # Test adding a file from S3
        app.add_data("s3://ai-infinitesearch")

        matched_images = app.get_data("A monument", [MEDIA_TYPE.IMAGE])

        # Verify that the file was added to the llm model
        self.assertEqual(["s3://ai-infinitesearch/test/building.jpeg"], matched_images)

    def test_audio_add_data(self):
        app = App(None, None, None)
        ChromaDB().reset()
        # Test adding a file from S3
        app.add_data(
            "s3://ai-infinitesearch/WhatsApp Ptt 2023-11-02 at 10.47.56.ogg",
            EmbeddingModelsConfig(),
            ChromaDB(),
        )

        matched_files = app.query(
            "s3", [MEDIA_TYPE.AUDIO], EmbeddingModelsConfig(), ChromaDB()
        )

        # Verify that the file was added to the llm model
        self.assertEqual(
            ["s3://ai-infinitesearch/WhatsApp Ptt 2023-11-02 at 10.47.56.ogg"],
            matched_files,
        )

    def test_add_data_with_nested_folders(self):
        utils = SourceUtils()
        db = ChromaDB()
        db.reset()

        # Test adding a file from S3
        utils.add_data("s3://ai-infinitesearch/test/b", Clip(), db, MEDIA_TYPE.IMAGE)

        matched_images = utils.query("A building", Clip(), db)

        # Verify that the file was added to the llm model
        self.assertEqual(
            [
                "s3://ai-infinitesearch/test/b/_9ea8f598-fdee-45b7-9338-46bce1d2f3a4.jpeg"
            ],
            matched_images,
        )



import uuid
from typing import Any

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from .base import BaseEmbeddingModel


class Clip(BaseEmbeddingModel):
    MODEL_NAME = "clip-ViT-B-32"
    SUPPORTED_MEDIA_TYPES = [MEDIA_TYPE.IMAGE]

    def __init__(self):
        self.model = None

    def get_media_encoding(
        self, data: Any, data_type: MEDIA_TYPE, datasource: DataSource
    ):
        """
        Applies the CLIP model to evaluate the vector representation of the supplied image
        """
        self._load_model()
        if data_type not in self.SUPPORTED_MEDIA_TYPES:
            raise ValueError(
                "Unsupported dataType. Clip model supports only {}".format(
                    self.SUPPORTED_MEDIA_TYPES
                )
            )
        image_features = self.model.encode(data)
        return {"embedding": [image_features.tolist()], "ids": [str(uuid.uuid4())]}

    def get_text_encoding(self, query: str):
        """
        Applies the CLIP model to evaluate the vector representation of the supplied text
        """
        self._load_model()
        text_features = self.model.encode(query)
        return {"embedding": text_features.tolist(), "meta_data": {}}

    def _load_model(self):
        """Load data from a director of images."""
        if self.model is None:
            from sentence_transformers import SentenceTransformer

            # load model
            self.model = SentenceTransformer(self.MODEL_NAME)

    def get_collection_name(self, media_type: MEDIA_TYPE):
        return "deepsearch-{}".format(media_type.name.lower())



from typing import Dict, List

from langchain.chat_models import ChatOpenAI
from langchain.schema import HumanMessage

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.types import MediaData, QueryResult
from .base import BaseLLM
from .configs.openai import OpenAiConfig


class OpenAi(BaseLLM):
    def __init__(self, config: OpenAiConfig = OpenAiConfig()):
        self.config = config
        super().__init__()

    def generate_prompt(self, input_query: str, contexts: List[str]) -> str:
        """
        Generates a prompt based on the given query and context, ready to be
        passed to an LLM

        :param input_query: The query to use.
        :type input_query: str
        :param contexts: List of similar documents to the query used as context.
        :type contexts: List[str]
        :return: The prompt
        :rtype: str
        """
        context_string = (" | ").join(contexts)
        # basic use case, no history.
        prompt = self.DEFAULT_PROMPT_TEMPLATE.substitute(
            context=context_string, query=input_query
        )
        return prompt

    def query(
        self,
        query: str,
        contexts: Dict[MEDIA_TYPE, List[MediaData]],
    ) -> QueryResult:
        results = []
        for item in contexts.items():
            media_data = item[1]
            for each_response in media_data:
                results.append(each_response.get("document", ""))
        prompt = self.generate_prompt(query, results)
        llm_response = self.get_llm_model_answer(prompt)
        query_result = {"llm_response": llm_response, "documents": contexts}
        return query_result

    def get_llm_model_answer(self, prompt) -> str:
        response = self._get_answer(prompt, self.config)
        return response

    def _get_answer(self, prompt: str, config: OpenAiConfig) -> str:
        messages = []
        messages.append(HumanMessage(content=prompt))
        kwargs = {
            "model": "gpt-3.5-turbo",
            "max_tokens": 1000,
            "model_kwargs": {},
        }
        chat = ChatOpenAI(**kwargs)
        return chat(messages).content



from string import Template
from typing import Dict, List

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.types import MediaData, QueryResult

DEFAULT_PROMPT = """
  Use the following pieces of context to answer the query at the end.
  If you don't know the answer, just say that you don't know, don't try to make up an answer.

  $context

  Query: $query

  Helpful Answer:
"""  # noqa:E501


class BaseLLM:
    DEFAULT_PROMPT_TEMPLATE = Template(DEFAULT_PROMPT)

    def __init__(self):
        pass

    def query(
        self,
        query: str,
        contexts: Dict[MEDIA_TYPE, List[MediaData]],
    ) -> QueryResult:
        raise NotImplementedError



import unittest

from deepsearchai.embedding_models.whisper_model import Whisper


class TestGetWhisperMediaEncoding(unittest.TestCase):
    def test_audio_encoding(self):
        """Test that the `get_media_encoding()` function can correctly identify the encoding of an audio file."""

        whisper = Whisper()
        # Load the audio file.
        audio_data = open("audio.ogg", "rb").read()

        # Get the media encoding.
        media_encoding = whisper.get_media_encoding(
            audio_data, whisper.MEDIA_TYPE.AUDIO
        )

        # Assert that the media encoding is correct.
        self.assertEqual(media_encoding, "audio/ogg;codecs=opus")

    def test_video_encoding(self):
        """Test that the `get_media_encoding()` function can correctly identify the encoding of a video file."""

        whisper = Whisper()
        # Load the video file.
        video_data = open("video.mp4", "rb").read()

        # Get the media encoding.
        media_encoding = whisper.get_media_encoding(
            video_data, whisper.MEDIA_TYPE.VIDEO
        )

        # Assert that the media encoding is correct.
        self.assertEqual(media_encoding, "video/mp4;codecs=h264")

    def test_invalid_media_type(self):
        """Test that the `get_media_encoding()` function raises a ValueError if an invalid media type is specified."""

        whisper = Whisper()
        # Get the media encoding.
        with self.assertRaises(ValueError):
            whisper.get_media_encoding(b"", "invalid_media_type")



import unittest
from unittest import mock
from unittest.mock import patch

import mock.mock

from deepsearchai.enums import MEDIA_TYPE
from deepsearchai.sources.data_source import DataSource
from deepsearchai.sources.local import LocalDataSource


class LocalDataSourceTest(unittest.TestCase):
    def setUp(self):
        self.local_data_source = LocalDataSource()

    @patch("os.walk")
    @patch("PIL.Image.open")
    def test_add_data_image_directory_with_no_existing_files(
        self, mock_image_file, mock_listdir
    ):
        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]

        # Create a mock image file
        image_data = mock.Mock()
        mock_listdir.return_value = [
            ("test_directory", "", ["image1.jpg", "image2.png"])
        ]
        mock_image_file.return_value = image_data

        # Create a mock vector database
        vector_database = mock.Mock()
        vector_database.get_existing_document_ids.return_value = []

        # Add local datasource data for a local directory
        directory = "test_directory"
        self.local_data_source.add_data(
            directory, embedding_models_config, vector_database
        )
        assert vector_database.add.mock_calls == [
            mock.call(
                image_data,
                DataSource.LOCAL,
                "test_directory/image1.jpg",
                directory,
                MEDIA_TYPE.IMAGE,
                embedding_model,
            ),
            mock.call(
                image_data,
                DataSource.LOCAL,
                "test_directory/image2.png",
                directory,
                MEDIA_TYPE.IMAGE,
                embedding_model,
            ),
        ]

    @patch("os.walk")
    @patch("PIL.Image.open")
    def test_add_data_image_directory_with_no_existing_files(
        self, mock_image_file, mock_listdir
    ):
        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]

        # Create a mock image file
        image_data = mock.Mock()
        mock_listdir.return_value = [
            ("test_directory", "", ["image1.jpg", "image2.png"])
        ]
        mock_image_file.return_value = image_data

        # Create a mock vector database
        vector_database = mock.Mock()
        vector_database.get_existing_document_ids.return_value = []

        # Add local datasource data for a local directory
        directory = "test_directory"
        self.local_data_source.add_data(
            directory, embedding_models_config, vector_database
        )
        assert vector_database.add.mock_calls == [
            mock.call(
                image_data,
                DataSource.LOCAL,
                "test_directory/image1.jpg",
                directory,
                MEDIA_TYPE.IMAGE,
                embedding_model,
            ),
            mock.call(
                image_data,
                DataSource.LOCAL,
                "test_directory/image2.png",
                directory,
                MEDIA_TYPE.IMAGE,
                embedding_model,
            ),
        ]

    @patch("os.walk")
    @patch("PIL.Image.open")
    def test_add_data_image_directory_with_existing_files(
        self, mock_image_file, mock_listdir
    ):
        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]

        # Create a mock image file
        image_data = mock.Mock()
        mock_listdir.return_value = [
            ("test_directory", "", ["image1.jpg", "image2.png"])
        ]
        mock_image_file.return_value = image_data

        # Create a mock vector database, such that one file already exists in it
        vector_database = mock.Mock()
        vector_database.get_existing_document_ids.return_value = [
            "test_directory/image2.png"
        ]

        # Add local datasource data for a local directory
        directory = "test_directory"
        self.local_data_source.add_data(
            directory, embedding_models_config, vector_database
        )
        assert vector_database.add.mock_calls == [
            mock.call(
                image_data,
                DataSource.LOCAL,
                "test_directory/image1.jpg",
                directory,
                MEDIA_TYPE.IMAGE,
                embedding_model,
            )
        ]

    @patch("os.path.isfile")
    @patch("PIL.Image.open")
    @patch("mimetypes.guess_type")
    def test_add_data_image(self, mock_mimetype, mock_image_file, mock_isfile):
        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]

        image_data = mock.Mock()
        mock_isfile.return_value = True
        mock_image_file.return_value = image_data
        mock_mimetype.return_value = "image/jpeg", "encoding"

        # Create a mock vector database, such that one file already exists in it
        vector_database = mock.Mock()
        vector_database.get_existing_document_ids.return_value = []
        directory = "image1.jpg"

        # Add local datasource data for a local directory
        self.local_data_source.add_data(
            directory, embedding_models_config, vector_database
        )
        assert vector_database.add.mock_calls == [
            mock.call(
                image_data,
                DataSource.LOCAL,
                directory,
                directory,
                MEDIA_TYPE.IMAGE,
                embedding_model,
            )
        ]

    @patch("os.path.isfile")
    @patch("mimetypes.guess_type")
    def test_add_data_audio(self, mock_mimetypes, mock_is_file):
        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]
        # Create a mock image file
        mock_is_file.return_value = True
        mock_mimetypes.return_value = "audio/mp3", "encoding"

        # Create a mock vector database, such that one file already exists in it
        vector_database = mock.Mock()
        vector_database.get_existing_document_ids.return_value = []

        filename = "audio.mp3"
        # Add local datasource data for a local directory
        self.local_data_source.add_data(
            filename, embedding_models_config, vector_database
        )
        assert vector_database.add.mock_calls == [
            mock.call(
                filename,
                DataSource.LOCAL,
                filename,
                filename,
                MEDIA_TYPE.AUDIO,
                embedding_model,
            )
        ]

    @patch("os.path.isfile")
    @patch("mimetypes.guess_type")
    def test_add_data_unsupported_media(self, mock_mimetype, mock_isfile):
        embedding_models_config = mock.Mock()
        embedding_model = mock.Mock()

        embedding_models_config.get_embedding_model.return_value = [embedding_model]
        # Create a mock unsupported file
        # Create a mock image file
        mock_isfile.return_value = True
        mock_mimetype.return_value = "random/format", "encoding"

        # Create a mock vector database, such that one file already exists in it
        vector_database = mock.Mock()
        vector_database.get_existing_document_ids.return_value = []

        # Create a mock for the llm model
        filename = "random.format"
        # Add local datasource data for a local directory
        self.local_data_source.add_data(
            filename, embedding_models_config, vector_database
        )
        vector_database.add.assert_not_called()
